( ******************************************* )
( ** NAXOS 2.1x NUM-LIBRARY                ** )
( ** Numeric routines, with 80x87 support  ** )
( ** Formatted numeric input and output    ** )
( ** Authors: K. Peper, A. Zissis          ** )
( ** Date : 7.8.92                         ** )
( ** Rev.2.1:  28/01/93                    ** )
( ******************************************* )

( ** Constants, Variables ** )

const round    $0000 ; ( Constants for NPU rounding control ) 
const rounddwn $0400 ;
const roundup  $0800 ;
const trunc    $0C00 ;
const dplc '.' ;         ( DecimalPointChar )

typ  FLOAT  Byte 8       ( Floating point var ) ;
typ  BCD    Byte 10      ( BCD-FLOAT var ) ;

var BASE    [ 10 ]       ( Base ) ;
var HLD     word         ( Pointer to Number processed ) ;
var DPL     word         ( Pointer to Decimalpoint ) ;
var LEADING [ *$20 ]     ( Leading Zero oder Space ) ;
var NUMBUF  byte 64      ( Conversion buffer for . d. etc. ) ;
var SGN     word         ( Sign ) ;

var FBUF    bcd          ( Conversion buffer for F. ) ;
var FDPL    word         ( Pointer to decimal point / FLOAT ) ;
var SCALE   float        ( scaling factor / FLOAT ) ;
 
 
( **  fstack operators ** )
MACRO F+     [ $9B $DE $C1 ]   ( FADDP ST(1),ST ) ;
MACRO F-     [ $9B $DE $E9 ]   ( FSUBP ST(1),ST ) ;
MACRO F*     [ $9B $DE $C9 ]   ( FMULP ST(1),ST ) ;
MACRO F/     [ $9B $DE $F9 ]   ( FDIVP ST(1),ST ) ;
MACRO FSQRT  [ $9B $D9 $FA ]   ( FSQRT ) ;
MACRO FABS   [ $9B $D9 $E1 ]   ( FABS ) ;
MACRO FNEG   [ $9B $D9 $E0 ]   ( FCHS ) ;
MACRO FDUP   [ $9B $D9 $C0 ]   ( FLD ST ) ;
MACRO FSWAP  [ $9B $D9 $C9 ]   ( FXCH ) ;
MACRO FOVER  [ $9B $D9 $C1 ]   ( FLD ST(1) ) ;
MACRO FSQR   FDUP F* ;
MACRO F0     [ $9B $D9 $EE ]   ( FLDZ )   ;
MACRO F1     [ $9B $D9 $E8 ]   ( FLD1 ) ;
MACRO FPI    [ $9B $D9 $EB ]   ( FLDPI ) ;
MACRO Fln    [ $D9 $ED $9B $D9 $C9 $9B $D9 $F1 ] ;
MACRO Flog   [ $D9 $EC $9B $D9 $C9 $9B $D9 $F1 ] ;


( ** memory to fstack    ** )
( ** memory_adress in SI ** )
MACRO M>F     [ $9B $DD $04 ]  ( FLD [si] )    ; (  8 bytes )
MACRO W/M>F   [ $9B $DF $04 ]  ( FILD w/[si] ) ; (  2 bytes )
MACRO D/M>F   [ $9B $DB $04 ]  ( FILD d/[si] ) ; (  4 bytes )
MACRO BCD/M>F [ $9B $DF $24 ]  ( FBLD [si] )   ; ( 10 bytes )

( ** fstack to memory    ** )
( ** memory_adress in SI ** )
MACRO F>M     [ $9B $DD $1C ]  ( FSTP [si] )    ; (  8 bytes )
MACRO W/F>M   [ $9B $DF $1C ]  ( FISTP w/[si] ) ; (  2 bytes )
MACRO D/F>M   [ $9B $DB $1C ]  ( FISTP d/[si] ) ; (  4 bytes )
MACRO BCD/F>M [ $9B $DF $34 ]  ( FBSTP [si] )   ; ( 10 bytes )

( ** register AX <-> fstack ** )
MACRO F>A     ( ST --> ax )     FBUF W/F>M m>a  ;
MACRO A>F     ( ax --> ST )     FBUF A>M W/M>F  ;

( ** register BX:AX <-> fstack ** )
: F>BA        ( ST --> bx:ax )  FBUF D/F>M m>ba ;
: BA>F        ( bx:ax --> ST )  FBUF ba>m D/M>F ;


( ** fconditions ** )
( CF:=1 if true, CF:=0 if false )

: .F=. [ $9B $DE $D9 $9B $DD $3E NPUSTAT $9B ]
  [ $50 $A1 NPUSTAT $25 $00 $45 $3D $00 $40 $58 $F8 $75 $01 $F5 ] ;

: .F<. [ $9B $DE $D9 $9B $DD $3E NPUSTAT $9B ]
  [ $50 $A1 NPUSTAT $25 $00 $45 $3D $00 $01 $58 $F8 $75 $00 $F5 ] ;

: .F>. [ $9B $DE $D9 $9B $DD $3E NPUSTAT $9B $50 $A1 NPUSTAT $25 $00 $45 ]
  [ $3D $00 $00 $58 $F8 $75 $01 $F5 ] ;

( *** miscellanelous NPU support functions *** )

: setdecimals 
( number of digits after decpoint )
( ******************************* )
( arg: BX = number of digits      )
( out: -                          )
( chg: AX                         )
( ******************************* )
  bpush
  , 18 - fdpl a>m
  %1 
  apop apush 1 do 
     %10 F*
  loop
  scale F>M
    apop 0 .=. IF %1 scale F>M ENDIF
;

: getdecimals 
( get number of digits after decpoint )
( ************************************* )
( arg: -                                )
( out: AX = number of digits            )
( chg: AX,BX                            )
( ************************************* )
  fdpl m>b , 18 -
;

: NPUMODE  
( set NPU rounding mode )
( ************************************* )
( arg: BX = rounding mode constant      )
( out: -                                )
( ************************************* )
  apush bpush adrpush epush 
  b>e
  [ $9B $D9 $3E NPUSTAT $9B ]
  NPUSTAT m>a $F3FF and e>b or NPUSTAT a>m  
  [ $9B $D9 $2E NPUSTAT $9B ]
  epop adrpop bpop apop
;


( ***  formatted output of BX:AX   *** )
( conversion according to current base )
( The complete series of <# ....... #> )
( has the following register effect:   )
( ************************************ )
( arg: BX:AX = double                  )
( out: SI:= string_adr                 )
( chg: CF?                             )
( ************************************ )


	   : hold 
( ** store char in numbuf ** )
( ************************** )
( arg: BX = char c           )
( out: SI:=SI-1              )
( chg: -                     )
( ************************** ) 
hld mem-     ( dekrement hld )
m>adr        ( get adress    )
bl>m         ( store char    )
;

	    : <# 
( ** start formatted output ** )
( **************************** )
( arg: BX:AX = unsigned double )
( out: ES:DI:= unsigned double )
( chg: -                       )
( **************************** )
fbuf ba>m 
[ $83 $C6 $04 ]    ( add si,04 )
ed>m
ba>ed          ( copy to ES:DI )
[ $BA numbuf ] ( mov dx,numbuf )
[ $81 $C2 *63 ]    ( add dx,63 )
[ $89 $16 hld ] ( mov [hld],dx )
;

: # 
 ( => ? )
     ( ud => ud' )
     ( => ? )
( ** eine Ziffer ablegen ** )
ed>ba 
.ba0=.                 ( auf Null prfen )
IF- NC
    base b/m>d         ( Basis nach DI )
	( ba/d: Ziffer berechnen )
	[ $93 ]      ( xchg ax,bx   ) 
	[ $31 $D2 ]  ( xor  dx,dx   ) 
	[ $F7 $F7 ]  ( div  di      ) ( (0:high)/i )
	[ $93 ]      ( xchg ax,bx   )
	[ $F7 $F7 ]  ( div  di      ) ( (rest:low)/i )
	[ $89 $D7 ]  ( mov  di,dx   ) ( rest )
    b>e xad            ( Zahl in es:di, Rest in ax )
    [ $05 $30 0 ]      ( add ax,30 : in ASCII-Format bringen )
    [ $3D $39 0 ]      ( cmp ax,39 : Zahlenbasis grer 10 ? )
    IF- NG
      [ $05 7 0 ]      ( add ax,07 : ja )
    ENDIF
ELSE
    leading m>a        ( Leerzeichen oder Null ausgeben )
ENDIF
a>b hold               ( Ziffer ablegen )
;


: #s  ( ud --> ud' )
( ** restliche Ziffern ablegen ** )
  begin
    ed>ba .ba0=.  ( Solange Zahl ungleich Null ... )
  while- NC
    #             ( ... Ziffer ausgeben )
  repeat
;


	    : #> 
( *** ends formatted output *** )
( ** string is ready to print * )
( ***************************** )
( arg: ES:DI double             )
( out: SI:= string_adr          )
( chg: SI,AX,BX                 )
( ***************************** )
adrpush
numbuf          
[ $81 $C6 *62 ] ( add si,62     )
[ $80 $3C $20 ] ( cmp [si],$20  )
IF- Z
[ $C6 $04 $30 ] ( mov b/[si],$30 )
ENDIF
fbuf m>ba
[ $83 $C6 $04 ]     ( add si,04 )
m>ed
adrpop
;


	    : sign 
( **** insert sign if n neg *** )
( ***************************** )
( arg: -                        )
( out: -                        )
( chg: BX,?                     )
( ***************************** )
  sgn 
  [ $D0 $64 1 ]  ( sal b/[si+1],1 : Vorzeichen negativ ? )
  if
    '-'   ( ja, dann Minuszeichen )
  else
    $20   ( nein, dann Leerzeichen )
  endif
  hold    ( Zeichen ausgeben )
;


: #f  ( FLOAT-Zahl in formatierte Ausgabe )
  scale m>f F*
  FBUF BCD/F>m
  0 , 8 DO
    i FBUF adrn+ m>a $0F and $30 + a>b hold
    i a>b , 8 - 2* 1+ FDPL m>b .=. IF DPLC hold ENDIF
    i FBUF adrn+ m>a u/2 u/2 u/2 u/2 $0f and $30 + a>b hold
    i a>b , 8 - 2* FDPL m>b .=. IF DPLC hold ENDIF
  LOOP
;

: FSIGN
  FBUF BCD/M>F
  F0 .F>. IF '-' hold ELSE $20 hold ENDIF
;

: f.  ( FLOAT-Zahl ausgeben )
  trunc NPUMODE
  FDUP 
  <# #f fsign #> a>d adr>e m>al apush
  FLOG F>A 0 .<. IF , 0 ENDIF fdpl m>b xab -
  e>adr ADRN+ apop adr- al>m adr>e
  d>a + 1+ a>d fdpl m>b - e>adr $.
  round NPUMODE
;

	  : d. 
( **** type double **** )
( ********************* )
( arg: BX:AX = double d )
( out: -                )
( chg: ?                )
( ********************* )
sgn b>m   ( save sign   )
d/abs     ( get absolut ) 
<# # #s   ( type double )
sign #>   ( add sign    )
adrpush
  sgn 
  [ $D0 $64 1 ]  ( sal b/[si+1],1 : Vorzeichen negativ ? )
  IF D/neg ENDIF
  adrpop
$.        ( and type    )
;

	 : ud. 
( type unsigned double  )
( ********************* )
( arg: BX:AX = double d )
( out: -                )
( chg: ?                )
( ********************* )
<# # #s #>
$.
;

	 : d.r 
( ****   type double   **** )
( r least signifdigits only )
( ************************* )
( arg: BX:AX = double d     )
(      DI = digits r        )
( out: -                    )
( chg: SI                   )
( ************************* )
  sgn b>m        ( save sign )
  d/abs          ( Absolutwert bilden )
  [ $51 ]        ( push cx )
  [ $89 $F9 ]    ( mov cx,di )
  <#             ( Start Ausgabe )
   FOR-
     #           ( Ausgabe entsprechend Feldbreite )
   NEXT-
  sign #>        ( Vorzeichen anfgen )
  adrpush
    sgn 
  [ $D0 $64 1 ]  ( sal b/[si+1],1 : Vorzeichen negativ ? )
  IF D/neg ENDIF
  adrpop
  [ $59 ]        ( pop cx )
  $.             ( ... und drucken )
;


	 : ud.r 
( ** type unsigned double * )
( r least signifdigits only )
( ************************* )
( arg: BX:AX = double d     )
(      DI = digits r        )
( out: -                    )
( chg: SI                   )
( ************************* )
  [ $51 ]        ( push cx )
  [ $89 $F9 ]    ( mov cx,di )
  <#             ( Start Ausgabe )
   FOR-
     #           ( Ausgabe entsprechend Feldbreite )
   NEXT-
  #>
  [ $59 ]        ( pop cx )
  $.             ( ... und drucken )
;



	   MACRO . 
( ***  type word in AX *** )
( ************************ )
( arg: AX = value n        )
( out: -                   )
( chg: BX                  )
( ************************ )
n>d   ( in doppelt genaue Zahl konvertieren )
d.    ( ... und ausgeben )
;


	   : .r 
( ***  type word in AX *** )
( r least sign.digits only )
( ************************ )
( arg: AX = value n        )
(      BX = digits r       )
( out: -                   )
( chg: -                   )
( ************************ )
bpush
b>d n>d   ( in doppelt genaue Zahl konvertieren )
d.r       ( ... und ausgeben )
bpop
;


	   : u. 
( type value in AX w/o sign )
( ************************* )
( arg: AX = value u         )
( out: -                    )
( chg: ?                    )
( ************************* )
 0       ( auf doppelte Genauigkeit erweitern )
<# # #s #> ( Konvertierung ohne Vorzeichen )
$.       ( ... und ausgeben )
;


	  : u.r 
( type value in AX w/o sign )
( r least sign.digits only  )
( ************************* )
( arg: AX = value u         )
(      BX = digits r        )
( out: -                    )
( chg: ?                    )
( ************************* )
[ $51 ]        ( push cx )
[ $89 $D9 ]    ( mov cx,bx )  
0              ( Absolutwert bilden )
<#             ( Start Ausgabe )
   FOR-
     #         ( Ausgabe entsprechend Feldbreite )
   NEXT-
#>             ( End         )
[ $59 ]        ( pop cx      )
$.             ( ... und drucken )
;

: .bit 
( ** type msb of AX ** )
( ******************** )
( arg: AX = value n    )
( out: AX = 2*n        )
( chg: AX              )
( ******************** )
[ $D1 $E0 ] ( shl ax,1 )
IF '1' else '0' endif
emit
;

	: u.b 
( ** type AX binary ** )
( ******************** )
( arg: AX = value n    )
( out: -               )
( chg: -               )
( ******************** )
apush bpush
.bit .bit .bit .bit space
.bit .bit .bit .bit space
.bit .bit .bit .bit space
.bit .bit .bit .bit space
bpop  apop
;

      macro lead0
( lead form. out by '0' )
( ********************* )
( arg: -                )
( out: -                )
( chg: -                )
( ********************* )
[ $C6 $06 leading $30 ] ( mov [leading],$30 )
;

      macro -lead0
( lead form. out by ' ' )
( ********************* )
( arg: -                )
( out: -                )
( chg: -                )
( ********************* )
[ $C6 $06 leading $20 ] ( mov [leading],$20 )
;


	 MACRO hex 
( ** set base to hex ** )
( ********************* )
( arg: -                )
( out: -                )
( chg: -                )
( ********************* )
[ $C6 $06 base $10 ] ( mov [si],16 )
;


	MACRO decimal 
( ** set base to decimal ** )
( ************************* )
( arg: -                    )
( out: -                    )
( chg: -                    )
( ************************* )
[ $C6 $06 base $0A ] ( mov [si],10 )
;

     MACRO .decdigit. 
( convert decdigit to number )
( ************************** )
( arg: AL = decdigit         )
( out: AL:= number           )
(      CF:=1 if digit valid  )
( chg: AL                    )
( ************************** ) 
[ $2C $30 ]      ( sub al,30 )
[ $72 $05 ]      ( jc ex     )
[ $3C $0A ]      ( cmp al,0A )
[ $73 $02 ]      ( jnc ex+1  )
[ $F5 ]          ( cmc       )
[ $F5 ]          ( ex: cmc   )
;

       proc .digit. 
( * convert digit to number  )
( valid_digit according base )
( ************************** )
( arg: AL = digit            )
( out: AL:= number           )
(      CF:=1 if digit valid  )
( chg: AL                    )
( ************************** ) 
   dpush
    '0' -
    16 .>.
   if
      7 -
   endif
    41 .>.
   if
      32 -
   endif
    base b/m>d d- 0 .>?<.
   dpop
;


var (base) [ 0 ]  ;
var (in)   [ 0 ]  ;

	: (number)
( * subroutine of number * ) 
( ************************ )
( arg: BP = string_adr     )
(      BX:AX = double      )
( out: BX:AX:= double'     )
( chg: ?                   )
( ************************ )
  begin
    [ $3E $8A $46 0 ]     ( mov al,ds:[bp+0] )
    [ $B4 0 ]             ( mov ah,0 : Zeichen holen )
    (in) b/mem-           ( Eingabe-Zhler dekrementieren )
    .digit.               ( prfen )
  while                   ( solange Ziffer )
    [ $45 ]               ( inc bp : adr +1  )
    apush
    ed>ba
    base b/m>d 
    ba*d                  ( Word mit Base multiplizieren )
    dpop
    E= 0 d/+            ( Ziffer aufaddieren )
    ba>ed
  repeat
;

	: number 
( convert string to double )
( var dpl counts digits... )
( after decimal point      )
( ************************ )
( arg: SI = string_adr     )
( out: BX:AX:= double      )
( chg: ?                   )
( ************************ )
  [ $55 ]               ( push bp )
  adr-
  m>al 1+               ( count holen )
  adr+
  [ $89 $F5 ]           ( mov bp,si )
  (in)  a>m             ( count speichern )
  base to (base) copy   ( Zahlenbasis retten )
  [ $3E $8A $46 0 ]     ( mov al,ds:[bp] )
  [ $B4 0 ]             ( mov ah,0 : 1.Zeichen )
  sgn a>m               ( merken ) 
  case
   '-' of               ( negative Zahl ? )
       (in) b/mem- [ $45 ]     ( dec (in) inc bp )
	;;
   '$' of
       (in) b/mem-            ( dec (in) )
       hex [ $45 ]            ( inc BP )
  endcase
  
  E= 0 D= 0 
  dpl d>m              ( Dezimalpunkt lschen )
  (number)             ( konvertieren )
  [ $3E $8A $46 0 ]    ( mov al,ds:[bp+0] )
  [ $B4 0 ]            ( mov ah,0 : Zeichen holen )
  case
   '.' of              ( Dezimalpunkt ? )
    (in) m>al 
    1- dpl al>m        ( Position merken )
    [ $45 ]            ( inc bp : nchste Adresse )
    (number) ;;        ( weiter konvertieren )
  endcase
  ed>ba  
  (base) to base b/copy  ( Basis wiedereinsetzen )
  sgn
  [ $8A $14 ]          ( mov dx,[si] : Vorzeichen holen )
  [ $80 $FA $2D ]      ( cmp dl,'-' )
  [ $F8 ]              ( clc )
  [ $75 $01 ]          ( jne +1 )
  [ $F5 ]              ( cmc )
   if d/neg endif      ( evtl. Vorzeichenwechsel )
  [ $5D ]              ( pop bp )
;


	: .number. 
( convert string to double )
(   with errorcode in CF   )
( var dpl counts digits... )
(   after decimal point    )
( ************************ )
( arg: SI = string_adr     )
( out: BX:AX:= double      )
(      CF:=1 if valid      )
( chg: ?                   )
( ************************ )
  number 
  (in) m>d
  [ $85 $FF ] ( test di,di )
  [ $F8 ]     ( clc )
  [ $75 $01 ] ( jnz 01     )
  [ $F5 ]     ( cmc        )
;

seal (base)


( ** formatted output of floating numbers ** )

: F>$ 
  trunc NPUMODE  
  adrpush
  FDUP 
  <# #f fsign #> a>d adr>e m>al apush
  FLOG F>A 0 .<. IF , 0 ENDIF fdpl m>b xab -
  e>adr ADRN+ apop adr- al>m adr>e
  d>a + 1+ a>d fdpl m>b - 
  adrpop al>m [ $89 $C1 ] adr>d e>adr ds>e [ $F3 $A4 ]
  round NPUMODE
;

: NPUERR
  " NAXOS 2.01: No 80x87 NPU detected, program aborted." $. 7 emit cr
  halt
;

macro initnoNPU init ;

: init overload ( - )
  ( !!! this init overloads the init method imported from KERN !!! )
  ~init
  [ $9B $DB $E3 ] ( FINIT )
  %12345 %12345 .F=. IF- NC NPUERR ENDIF
  round NPUMODE  
  3 setdecimals
;



var (#in) "                                              " ;

	  proc #in 
( * input double         ** )
( **   to current base   ** )
( ************************* )
( arg: BX = digitcountmax   )
(      ES:DI = default d    )
( out: BX:AX:= double|def   )
( chg: ?                    )
( ************************* )
begin
   (#in) expect
   (#in) adr- m>al .0=. if 
       ed>ba exit 
   endif
   (#in) .number. .not.
while
 7 emit
repeat
;

: .#Fin.
  21 #in BA>F dpl m>a 0 .<>.
  IF dpl m>a 1 do %10 F/ loop
  ENDIF
;
