(
	************************************************************
	*         NAXOS Version 2.13  16-Bit Runtime library       *
	*                                                          *
	*         VGA 256-COLOR MODE GRAPHIC ROUTINES V1.02 ET4000 *
	*                                                          *
	* (C) 1989,90,91,92,93 by K.Peper, A.Zissis, I.Tossounidis *
	*                                                          *
	*                      Rev. 13B 25.08.93                   *
	*                                                          *
	************************************************************
)



( Standard Modes available on all VGA )
(             ModeName     *$nnxx / nn=Devcaps# xx=Mode# )
	
	const 320x200x256  *$0213 ; ( 320x200 Pixel 256 color VGA/MCGA Mode )
	const 80x25x16     *$1503 ; ( VGA-16 color textmode )


( ---------------------------------------------------------------- )
(              Start of the card dependent patch Area              )
(              Tseng ET4000 Chipset Rev. 1.02 02.04.93             )
( ---------------------------------------------------------------- )

( SVGA modes on ET4000 based systems )
(             ModeName     *$nnxx / nn=Devcaps# xx=Mode# (depends on card) )
	
	const 640x350x256  *$092D ;
	const 640x400x256  *$0A2F ;
	const 640x480x256  *$0B2E ;
	const 800x600x256  *$0D30 ;
	const 1024x768x256 *$0F38 ;

	const 640x480x32K  *$102E ;
	const 800x600x32K  *$1130 ;
	const 640x480x64K  *$122E ;
	const 800x600x64K  *$1330 ;
	const 640x480x16M  *$142E ;

( Bankswitch routines for HI-RES 256color modes )
( This routines depend on the chipset used by the )
( VGA-Adaptor )

	macro wbanksel ( ET3000/ET4000 only ! )
	( select 64K video page for write access )
	( arg : AX := PAGE to write into )
	( out : -               )
	( chg : -               )
	[ $BA $3CD $EE ] ( mov dx,03CDH , out dx,al )
	;

	macro rbanksel ( ET3000/ET4000 only ! )
	( select 64K video page for read access )
	( arg : AX := PAGE to read from )
	( out : -               )
	( chg : CX              )
	[ $B1 $04        ( mov cl,4 )
	  $D3 $E0        ( shl ax,cl )
	  $BA $3CD       ( mov dx,03CDH )
	  $EE ]          ( out dx,al )
	;

(   devcaps  x-res y-res pages colors / all starting at zero )

var devcaps [ *319 *199 00 003
	      *319 *199 07 015 
	      *319 *199 00 255 
	      *639 *199 00 001
	      *639 *199 03 015
	      *639 *349 01 004
	      *639 *349 01 015
	      *639 *479 00 001
	      *639 *479 00 015
	      *639 *350 00 255
	      *639 *399 00 255
	      *639 *479 00 255        
	      *799 *599 00 015
	      *799 *599 00 255
	     *1023 *767 00 015        
	     *1023 *767 00 255 
	      *639 *479 00 255
      	      *799 *599 00 015
	      *639 *479 00 255
	      *799 *599 00 015
	      *639 *479 00 255
	       *79 *024 07 015 ] ;

( ---------------------------------------------------------------- )
(                End of the card dependent patch Area              )
( ---------------------------------------------------------------- )

var pix-x-max [ *79 ] ;
var pix-y-max [ *23 ] ;
var page-max  [ 8 ] ;
var color-max [ *15 ] ;

var page      [ 0 ] ;   
var pix-x     [ *0 ] ;
var pix-y     [ *0 ] ;

var color     [ $FF ] ;

: setmode ( switch to graphics mode specified in bl -> mode constants )
  b>a $10 intr
;

: setmodenc ( as above with no clear screen )
  b>a $80 or $10 intr
;

: graphmode ( switch to graphicsmode / input: mode.const in BX ) 
  bpush B>a split b>a 6 * devcaps adrn+ to pix-x-max , 6 i/copy
  apop split $10 intr 
  pix-y-max ( this line was added to suit var-usage by optimizing V2.14 compiler )
;

: graphmode-16M ( switch to graphicsmode / input: mode.const in BX ) 
  bpush B>a split b>a 6 * devcaps adrn+ to pix-x-max , 6 i/copy
  apop split , $22e $10 intr 
;

: graphmodenc ( switch to graphicsmode / input: mode.const in BX ) 
  bpush B>a split b>a devcaps adrn+ to pix-x-max , 6 i/copy
  apop split $80 or $10 intr pix-y-max
;

: textmode ( switch to standard 80x25 Textmode )
  80x25x16 graphmode  
;

: textmodenc ( as above with no clear screen )
  80x25x16 graphmodenc  
;


: entrymode ( restore initial videomode as on program entry )
  video-byte m>bl setmode
;

: setcolor ( set pixel color as in bx for 256Color Modes )
  color bl>m 
;

: disp-page ( select video page to display as specified in bl )
  b>a $5 combine $10 intr
;

: plot-page ( select video page to plot into as spec. in bl )
  page bl>m
;

: putpixel ( x-coord in bx; y-coord in ax; color is taken from pix-color )
  e= $A000
  [ $89 $1E pix-x ]  ( mov pix-x,bx )
  [ $A3 pix-y ]      ( mov pix-y,ax )
  xab  a>d
  [ $A1 pix-x-max ] 1+ *
  [ $87 $DF ]        ( xchg BX,DI )
   + [ $83 $D7 $00 ] ( adc di,0 )
   a>adr
  d>a wbanksel
  [ $A0 color ]    ( mov al,color )
  es: al>m
;
  
: getpixel ( x-coord in bx; y-coord in ax; color is returned in al )
  ipush
  e= $A000
  [ $89 $1E pix-x ]  ( mov pix-x,bx )
  [ $A3 pix-y ]      ( mov pix-y,ax )
  xab  a>d
  [ $A1 pix-x-max ] 1+ *
  [ $87 $DF ]        ( xchg BX,DI )
   + [ $83 $D7 $00 ] ( adc di,0 )
   a>adr
  d>a rbanksel
  es: m>al
  ipop
;

: hline ( x1-coord in bx;y coord in ax; x2-coord in di )
  putpixel
  d>a pix-x m>b - a>b FOR
    pix-x m>b b+ b>m pix-y m>a putpixel
  next
;


: vline ( x1-coord in bx;y coord in ax; y2-coord in di )
  putpixel
  d>a pix-y m>b - a>b FOR
    pix-x m>b pix-y m>a 1+ a>m putpixel
  next
;

var BG-x1 word ;
var BG-y1 word ;
var BG-xd word ;
var BG-yd word ;
var BG-xf word ;




: line  ( input: x1: BX, y1: AX, x2: ES, y2: DI )
	( output: - )
	( CX is modified ! )
	
[ $89 $DA ] d>b
[ $89 $1E BG-y1 $E8 $96 $00 $89 $0E BG-yd $E8 $AE ]
[ $00 $92 ]
e>b
[ $89 $1E BG-x1 $E8 $85 $00 $89 $0E BG-xd $E8 $9D ]
[ $00 $8B $D8 $3B $DA $7E $3C $8B $C2 $03 $C0 $2B $C3 $A3 BG-xf ]
[ $8B $CB $41 $E8 $75 $00 $A1 BG-xf $0B $C0 $7E $15 $03 $C2 $03 ]
[ $C2 $2B $C3 $2B $C3 $A3 BG-xf $A1 BG-yd $01 $06 BG-y1 $EB ]
[ $08 $90 $03 $C2 $03 $C2 $A3 BG-xf $A1 BG-xd $01 $06 BG-x1 ]
[ $E2 $D1 $C3 $8B $C3 $03 $C0 $2B $C2 $A3 BG-xf $8B $CA $41 $E8 ]
[ $39 $00 $A1 BG-xf $0B $C0 $7E $15 $03 $C3 $03 $C3 $2B $C2 $2B ]
[ $C2 $A3 BG-xf $A1 BG-xd $01 $06 BG-x1 $EB $08 $90 $03 $C3 ]
[ $03 $C3 $A3 BG-xf $A1 BG-yd $01 $06 BG-y1 $E2 $D1 $C3 $33 ]
[ $C9 $2B $C3 $74 $05 $78 $02 $41 $C3 $49 $C3 $52 $53 $50 $8B ]
[ $1E BG-x1 $A1 BG-y1 ]
 putpixel
[ $58 $5B $5A $C3 ]
[ $0B $C0 $79 $02 $F7 $D8 $C3 ]
;


: lineto ( x in BX , y in AX )
  [ $06 $51 ] ( push es , push cx )
  bpush apush
  a>d b>e
  pix-x m>b pix-y m>a
  line
  apop pix-y a>m apop pix-x a>m
  [ $59 $07 ] ( pop cx , pop es )
;
