(
        ************************************************************
        *         NAXOS Version 2.13  16-Bit Runtime library       *
        *                                                          *
        *                 iAPX386/486 Support V1.01                *
        *                                                          *
        * (C) 1989,90,91,92,93 by K.Peper, A.Zissis, I.Tossounidis *
        *                                                          *
        *                      Rev. 13B 25.08.93                   *
        *                                                          *
        ************************************************************
)

(  ************************************************************** 
   *								*
   * Attention: All XM operation need XMEM386 to be installed	*
   *								*
   ************************************************************** )
   
( Syntax:       xa = eax   = 32Bit a
		xb = ebx   = 32Bit b
		xi = ecx   = 32Bit i
		xd = edi   = 32Bit d
		xadr = esi = 32Bit adr
		xm = memory cell in 4G adr.space
		xe* = extended part [upper 16 Bit] of *
		xea = upper 16 Bit of a
		...
		ux  = unsigned 32Bit value 
		x/  = 32Bit Prefix
		ux/ = unsigned 32Bit Prefix )
		

var msw [ *0 ] ; ( machine-status-word: Bit0=0 eq Real mode; Bit0=1 prot. )

( Prefixes )

macro FS: [ $64 ] ;
macro GS: [ $65 ] ;

( Convert signed reg16 to signed reg32  )

macro a>xa [ $66 $0F $BF $C0 ] ;        ( movsx eax,ax )
macro b>xb [ $66 $0F $BF $DB ] ;        ( movsx ebx,bx )
macro i>xi [ $66 $0F $BF $C9 ] ;        ( movsx ecx,cx )
macro d>xd [ $66 $0F $BF $FF ] ;        ( movsx edi,di )
macro adr>xadr [ $66 $0F $BF $F6 ] ;    ( movsx esi,si )

( Convert unsigned reg16 to unsigned reg32  )

macro ua>xa [ $66 $0F $B7 $C0 ] ;       ( movzx eax,ax )
macro ub>xb [ $66 $0F $B7 $DB ] ;       ( movzx ebx,bx )
macro ui>xi [ $66 $0F $B7 $C9 ] ;       ( movzx ecx,cx )
macro ud>xd [ $66 $0F $B7 $FF ] ;       ( movzx edi,di )
macro uadr>xadr [ $66 $0F $B7 $F6 ] ;   ( movzx esi,si )

( move reg32 to reg32  )

macro xb>xa [ $66 $8B $C3 ] ;           ( mov eax,ebx )
macro xi>xa [ $66 $8B $C1 ] ;           ( mov eax,ecx )
macro xd>xa [ $66 $8B $C7 ] ;           ( mov eax,edi )
macro xadr>xa [ $66 $8B $C6 ] ;         ( mov eax,esi )

macro xa>xb [ $66 $8B $D8 ] ;           ( mov ebx,eax )
macro xi>xb [ $66 $8B $D9 ] ;           ( mov ebx,ecx )
macro xd>xb [ $66 $8B $DF ] ;           ( mov ebx,edi )
macro xadr>xb [ $66 $8B $DE ] ;         ( mov ebx,esi )

macro xa>xd [ $66 $8B $F8 ] ;           ( mov edi,eax )
macro xb>xd [ $66 $8B $FB ] ;           ( mov edi,ebx )
macro xi>xd [ $66 $8B $F9 ] ;           ( mov edi,ecx )
macro xadr>xd [ $66 $8B $FE ] ;         ( mov edi,esi )

macro xa>xadr [ $66 $8B $F0 ] ;         ( mov esi,eax )
macro xb>xadr [ $66 $8B $F3 ] ;         ( mov esi,ebx )
macro xi>xadr [ $66 $8B $F1 ] ;         ( mov esi,ecx )
macro xd>xadr [ $66 $8B $F7 ] ;         ( mov esi,edi )

macro xa>xi [ $66 $8B $C8 ] ;           ( mov ecx,eax )
macro xb>xi [ $66 $8B $CB ] ;           ( mov ecx,ebx )
macro xd>xi [ $66 $8B $CF ] ;           ( mov ecx,edi )
macro xadr>xi [ $66 $8B $CE ] ;         ( mov ecx,esi )

( Move upper 16Bit to lower 16Bit )

macro xea>a [ $66 $C1 $E8 $10 ] ;       ( shr eax,16 )
macro xeb>b [ $66 $C1 $EB $10 ] ;       ( shr ebx,16 )
macro xed>d [ $66 $C1 $EF $10 ] ;       ( shr edi,16 )
macro xeadr>adr [ $66 $C1 $EE $10 ] ;   ( shr esi,16 )

( Move doubles reg16:reg16 to reg32 )

: ba>xa 
  [ $8B $D3				( mov dx,bx )
    $66 $C1 $E2 $10			( shl edx,16 )
    $8B $D0				( mov dx,ax )
    $66 $8B $C2 ] 			( mov eax,edx )
;
 
: ba>xb 
  [ $8B $D3				( mov dx,bx )
    $66 $C1 $E2 $10			( shl edx,16 )
    $8B $D0				( mov dx,ax )
    $66 $8B $DA ] 			( mov ebx,edx )
;  


: ba>xi 
  [ $8B $D3				( mov dx,bx )
    $66 $C1 $E2 $10			( shl edx,16 )
    $8B $D0				( mov dx,ax )
    $66 $8B $CA ] 			( mov ecx,edx )
;  

: ba>xd 
  [ $8B $D3				( mov dx,bx )
    $66 $C1 $E2 $10			( shl edx,16 )
    $8B $D0				( mov dx,ax )
    $66 $8B $FA ] 			( mov edi,edx )
;


: ba>xadr
  [ $8B $D3				( mov dx,bx )
    $66 $C1 $E2 $10			( shl edx,16 )
    $8B $D0				( mov dx,ax )
    $66 $8B $F2 ] 			( mov esi,edx )
;



: ed>xa 
  [ $8C $C2				( mov dx,es )
    $66 $C1 $E2 $10			( shl edx,16 )
    $8B $D7				( mov dx,di )
    $66 $8B $C2 ] 			( mov eax,edx )
;

: ed>xb
  [ $8C $C2				( mov dx,es )
    $66 $C1 $E2 $10			( shl edx,16 )
    $8B $D7				( mov dx,di )
    $66 $8B $DA ] 			( mov ebx,edx )
;

: ed>xi
  [ $8C $C2				( mov dx,es )
    $66 $C1 $E2 $10			( shl edx,16 )
    $8B $D7				( mov dx,di )
    $66 $8B $CA ] 			( mov ecx,edx )
;


: ed>xd 
  [ $8C $C2				( mov dx,es )
    $66 $C1 $E2 $10			( shl edx,16 )
    $8B $D7				( mov dx,di )
    $66 $8B $FA ] 			( mov edi,edx )
;

: ed>xadr
  [ $8C $C2				( mov dx,es )
    $66 $C1 $E2 $10			( shl edx,16 )
    $8B $D7				( mov dx,di )
    $66 $8B $F2 ] 			( mov esi,edx )
;


( move doubles reg32 to reg16:reg16 )

macro xa>ba
  [ $66 $8B $D8				( mov ebx,eax )
    $66 $C1 $EB $10 ] ;			( shr ebx,16 )
    
macro xb>ba
  [ $8B $C3 				( mov ax,bx )
    $66 $C1 $EB $10 ] ;			( shr ebx,16 )    

macro xi>ba
  [ $8B $C1				( mov ax,cx )
    $66 $C1 $E9 $10			( shr ecx,16 )
    $8B $D9 ] ;				( mov bx,cx )
    
macro xd>ba
  [ $8B $C7				( mov ax,di )
    $66 $C1 $EF $10			( shr edi,16 )
    $8B $DF ] ;				( mov bx,di )
    
macro xadr>ba
  [ $8B $C6				( mov ax,si )
    $66 $C1 $EE $10			( shr esi,16 )
    $8B $DE ] ;				( mov bx,si )


macro xa>ed
  [ $8B $F8				( mov di,ax )
    $66 $C1 $E8 $10			( shr eax,16 )
    $8E $C0 ] ;				( mov es,ax )

macro xb>ed
  [ $8B $FB				( mov di,bx )
    $66 $C1 $EB $10			( shr ebx,16 )
    $8E $C3 ] ;				( mov es,bx )

macro xi>ed
  [ $8B $F9				( mov di,cx )
    $66 $C1 $E9 $10			( shr ecx,16 )
    $8E $C1 ] ;				( mov es,cx )


: xd>ed
  [ $8B $D7				( mov dx,di )
    $66 $C1 $EF $10			( shr edi,16 )
    $8E $C7 				( mov es,di )
    $8B $FA ] ;				( mov di,dx )


macro xadr>ed
  [ $8B $FE				( mov di,si )
    $66 $C1 $EE $10			( shr esi,16 )
    $8E $C6 ] ;				( mov es,si )


( Memory access )


macro m>xa [ $66 $8B $04 ] ;		( mov eax,[si] )
macro xa>m [ $66 $89 $04 ] ;		( mov [si],eax )
macro xm>a [ $67 $8B $06 ] ;		( mov ax,[esi] )
macro a>xm [ $67 $89 $06 ] ;		( mov [esi],ax )
macro xm>xa [ $67 $66 $8B $06 ] ;	( mov eax,[esi] )
macro xa>xm [ $67 $66 $89 $06 ] ;	( mov [esi],eax )

macro m>xb [ $66 $8B $1C ] ;		( mov ebx,[si] )
macro xb>m [ $66 $89 $1C ] ;		( mov [si],ebx )
macro xm>b [ $67 $8B $1E ] ;		( mov bx,[esi] )
macro b>xm [ $67 $89 $1E ] ;		( mov [esi],bx )
macro xm>xb [ $67 $66 $8B $1E ] ;	( mov ebx,[esi] )
macro xb>xm [ $67 $66 $89 $1E ] ;	( mov [esi],ebx )
 
macro m>xi [ $66 $8B $0C ] ;		( mov ecx,[si] )
macro xi>m [ $66 $89 $0C ] ;		( mov [si],ecx )
macro xm>i [ $67 $8B $0E ] ;		( mov cx,[esi] )
macro i>xm [ $67 $89 $0E ] ;		( mov [esi],cx )
macro xm>xi [ $67 $66 $8B $0E ] ;	( mov ecx,[esi] )
macro xi>xm [ $67 $66 $89 $0E ] ;	( mov [esi],ecx )

macro m>xd [ $66 $8B $3C ] ;		( mov edi,[si] )
macro xd>m [ $66 $89 $3C ] ;		( mov [si],edi )
macro xm>d [ $67 $8B $3E ] ;		( mov di,[esi] )
macro d>xm [ $67 $89 $3E ] ;		( mov [esi],di )
macro xm>xd [ $67 $66 $8B $3E ] ;	( mov edi,[esi] )
macro xd>xm [ $67 $66 $89 $3E ] ;	( mov [esi],edi )

macro m>xadr [ $66 $8B $34 ] ;		( mov esi,[si] )
macro xadr>m [ $66 $89 $34 ] ;		( mov [si],esi )
macro xm>adr [ $67 $8B $36 ] ;		( mov si,[esi] )
macro adr>xm [ $67 $89 $36 ] ;		( mov [esi],si )
macro xm>xadr [ $67 $66 $8B $36 ] ;	( mov esi,[esi] )
macro xadr>xm [ $67 $66 $89 $36 ] ;	( mov [esi],esi )


( Simple arithmetics )

macro x/1+ [ $66 $40 ] ;                ( inc eax )
macro x/1- [ $66 $48 ] ;                ( dec eax )
macro x/2+ x/1+ x/1+ ;
macro x/2- x/1- x/1- ;
macro x/4+ x/2+ x/2+ ;
macro x/4- x/2- x/2- ;

macro x/adr+ [ $66 $46 ] ;              ( inc esi )
macro x/adr- [ $66 $4E ] ;              ( dec esi )
macro x/adr2+ x/adr+ x/adr+ ;
macro x/adr2- x/adr- x/adr- ;
macro x/adr4+ x/adr2+ x/adr2+ ;
macro x/adr4- x/adr2- x/adr2- ;

macro x/+ [ $66 $03 $C3 ] ;             ( add eax,ebx )
macro x/- [ $66 $2B $C3 ] ;             ( sub eax,ebx )
macro x/* [ $66 $F7 $EB ] ;             ( imul ebx )
macro x/  [ $66 $F7 $FB ] ;             ( idiv ebx )
macro ux/* [ $66 $F7 $E3 ] ;            ( mul ebx )
macro ux/  [ $66 $F7 $F3 ] ;            ( div ebx )

macro x/2* [ $66 $B7 $C0 ] ;            ( add eax,eax )
macro ux/2/ [ $66 $C1 $E8 $01 ] ;       ( shr eax,1 )
macro ux/4/ [ $66 $C1 $E8 $02 ] ;       ( shr eax,2 )
macro ux/8/ [ $66 $C1 $E8 $03 ] ;       ( shr eax,3 )
macro ux/16/ [ $66 $C1 $E8 $04 ] ;      ( shr eax,4 )



: real-mode?
  [ $0F $01 $26 msw ]	( smsw W/[cr0] )
  msw [ $8B $04 ]    	( mov  ax,[si] )
  [ $25 $01 $00 ]    	( and  ax,0001 )   
  [ $85 $C0 ]		( test ax,ax )
  [ $75 $01 ]		( jnz +01 )
  [ $F5 ]		( cmc )
;


LABEL GO-REAL ;
csvar GDT-PTR  [ *$18 ] ;
csvar GDT-PTR2 [ &0 ] ;
csvar GDT [ 	&0 &0
		*$FFFF
		*0
		0
		$92
		$CF
		$0
	   ] ;
	   
csvar FARJMP  [ *0 ] ;	   
csvar FARJMP2 [ *0 ] ;


proc (XMEM386) 
  [ $8C $C8 			( mov ax,cs )
    $66 $0F $B7 $C0 		( movzx eax,ax )
    $66 $C1 $E0 $04 		( shl eax,4 )
    $BB GDT 			( mov bx,offset GDT )
    $66 $0F $B7 $DB 		( movzx ebx,bx )
    $66 $03 $C3			( add eax,ebx )
    $2E $66 $A3 GDT-PTR2	( mov D/cs:[GDT-PTR2],eax )
    $9C				( PUSHF )
    $1E				( PUSH DS )
    $06				( PUSH ES )
    $0F $A0			( PUSH FS )
    $0F $A8			( PUSH GS )
    $FA				( CLI )
    $2E $0F $01 $16 GDT-PTR	( LGDT CS:[GDT-PTR] )
    $0F $20 $C0			( mov eax,cr0 )
    $66 $83 $C8 $01		( or eax,1 )
    $0F $22 $C0			( mov cr0,eax )
  ( now in protected mode )
    $EB $00			( jmp short +$00 / clear prefetch queue )	
    $B8 *$8			( mov ax,$0008 )
    $8E $D8			( mov ds,ax / load selector 1 from GDT rpl=0 )
    $8E $C0			( mov es,ax / load selector 1 from GDT rpl=0 )
    $8E $E0			( mov fs,ax / load selector 1 from GDT rpl=0 )    	
    $8E $E8			( mov gs,ax / load selector 1 from GDT rpl=0 )
    $0F $20 $C0			( mov eax,cr0 )
    $66 $83 $E0 $FE		( and eax, not 1 )
    $0F $22 $C0 ]			( mov cr0,eax )
  ( now back to real mode ==> DS,ES,FS,GS segment limits are $FFFFFFFF now )	    
 label go-real
  [ $0F $A9			( pop GS )
    $0F $A1			( pop FS )
    $07				( pop ES )
    $1F				( pop DS )
    $9D				( popf )
  ]           
;	     

proc XMEM386
( *********************************
  * Switch iAPX386+ to special    *
  * real-mode with 4G- Address    *
  * space; doesn't work in V86 or *
  * protected mode environments   *
  * (Windows; OS/2; EMM386...)    *
  * arg: -			  *
  * out: -			  *
  * chg: all			  *
  ********************************* )


  real-mode? IF 

  [  $2E $8C $0E FARJMP2	( mov W/cs:[FARJMP2],cs )
     $2E $C7 $06 FARJMP GO-REAL	( mov W/cs:[FARJMP],offset GO-REAL )
  ]
  (XMEM386)
  ELSE
   $CD errorptr
  ENDIF
;  